
% ROOTCLUS main program 
% Written by L. Bocci and D. Vicari, Sapienza University of Rome, Italy (Last update: Apr, 2019)

function [P,Mh,W,Vh,Cost,f,F,iter,Ldif]=ROOTCLUS_main(S,s,P0,Mh0,W0,Cost0,offdiag,lambdal)


warning off all; 
optlsqnonneg=optimset('Display','off','TolFun',1e-6);

conv=1e-8;      

[N,J,H]=size(Mh0); 
nodiag=1-eye(N); oneN2=ones(N^2,1);

SS=ssq(s);
NP=sum(sum(P0));

        %% ------ INITIALIZATION: STARTING VALUES Vh0, Cost0

Qh=zeros(N^2,J,H);  
 for h=1:H, Qh(:,:,h)=khatrirao(P0,Mh0(:,:,h))+khatrirao(Mh0(:,:,h),P0+Mh0(:,:,h)); end
T=khatrirao(P0,P0); w0=diag(W0);               
res0=s-repmat(T*w0,1,H)-repmat(Cost0',N^2,1); 

 if offdiag==1
   for h=1:H, Qh(:,:,h)=Qh(:,:,h).*repmat(nodiag(:),1,J); end
   T=T.*repmat(nodiag(:),1,J);
   res0=res0.*repmat(nodiag(:),1,H);
 end
 
 Vh0=zeros(H,J);
if NP<N
    for h=1:H
        Vh0(h,:)=lsqnonneg(Qh(:,:,h)+T,res0(:,h),optlsqnonneg); 
    end
end
   sM=reshape(sum(Mh0,1),J,H)';
   Vh0=Vh0.*(sM~=0); 

 for h=1:H
      Resi=s(:,h)-T*w0-(T+Qh(:,:,h))*Vh0(h,:)';       
     if offdiag==1
        Cost0(h,1)=sum(Resi)/(N*(N-1));
     else
        Cost0(h,1)=sum(Resi)/(N*N);
     end
 end
 
 f=0;
if offdiag==1
    for h=1:H
        f=f+ssq(s(:,h)-T*w0-(T+Qh(:,:,h))*Vh0(h,:)'-Cost0(h)*oneN2.*nodiag(:));
    end
elseif offdiag==0
    for h=1:H
        f=f+ssq(s(:,h)-T*w0-(T+Qh(:,:,h))*Vh0(h,:)'-Cost0(h)*oneN2);
    end
end
f=f/SS + lambdal*(N-NP)/(NP^2);
fold=inf; iter=0;  

while (fold-f)>conv*f || iter<2  
 iter=iter+1; 
 fold=f;
  
           %% ------ STEP 1 Updating P and Mh
                                                      
 Resh=S-reshape(kron(Cost0',ones(N)),[N,N,H]);
 res_Cost=s-repmat(Cost0',N^2,1);   

 if offdiag==1 
     Resh=Resh.*repmat(nodiag,[1,1,H]); 
     res_Cost=res_Cost.*repmat(nodiag(:),1,H);
 end
 
 %%%%%
   sz=1; itP=0;
 while  (sum(sz(:))>0) 
     itP=itP+1; if (itP>3), break, end
     [P,Mh]=PMh_ROOTCLUS(SS,Resh,P0,Mh0,W0,Vh0,offdiag,lambdal);
     NP=sum(sum(P));
     whereP=(sum(P)==0);
     sM=reshape(sum(Mh,1),J,H)';  whereM=(sM>0);
     sz=whereM(:,whereP);
     P0=P; Mh0=Mh;  
     
 end
  %%%%%
 
 T=khatrirao(P,P);  Qh=zeros(N^2,J,H); 
 for h=1:H, Qh(:,:,h)=khatrirao(P,Mh(:,:,h))+khatrirao(Mh(:,:,h),P+Mh(:,:,h));  end
 if offdiag==1
     T=T.*repmat(nodiag(:),1,J);
     for h=1:H, Qh(:,:,h)=Qh(:,:,h).*repmat(nodiag(:),1,J); end
 end
 %%
           %% ------ STEP 2 Updating Vh 

  Vh=zeros(H,J); 
  if NP<N
    res=res_Cost-repmat(T*w0,1,H);     
    
    for h=1:H
        Vh(h,:)=lsqnonneg(T+Qh(:,:,h),res(:,h),optlsqnonneg);   
    end
        Vh=(sM>0).*Vh;    
  end
  
           %% ------ STEP 3 Updating W 

  Ah=zeros(N^2,H);  
  for h=1:H, Ah(:,h)=(T+Qh(:,:,h))*Vh(h,:)'; end  
  res=res_Cost-Ah; 
  
  W=diag(lsqnonneg(T,mean(res,2),optlsqnonneg));  

           %% ------ STEP 4 Updating Cost 

 Cost=zeros(H,1);    
 for h=1:H
     Resi=s(:,h)-T*diag(W)-(T+Qh(:,:,h))*Vh(h,:)';  
     if offdiag==1
        Cost(h,1)=sum(Resi)/(N*(N-1));
     else
        Cost(h,1)=sum(Resi)/(N*N);
     end
 end

 f=0;
 if offdiag==1
    for h=1:H
        f=f+ssq(s(:,h)-T*diag(W)-(T+Qh(:,:,h))*Vh(h,:)'-Cost(h)*oneN2.*nodiag(:));
    end
 else
    for h=1:H
        f=f+ssq(s(:,h)-T*diag(W)-(T+Qh(:,:,h))*Vh(h,:)'-Cost(h)*oneN2);
    end
 end
 f=f/SS + lambdal*(N-NP)/(NP^2);
 
  P0=P; Mh0=Mh; Vh0=Vh; W0=W; w0=diag(W); Cost0=Cost;

Ldif=fold-f;

end

F = f - lambdal*(N-NP)/(NP^2);   % non penalized loss function

end

%%%%%%%%%%%%%%%%%%%%%%
%%% NESTED FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%

%%%%%% Computing sum of squares
function t = ssq(a)
     t = sum(sum(a.^2));
end
%%%%%%

%%%%%%% Updating membership matrices P and Mh in Step 1 of ROOTCLUS_main  
% Written by L. Bocci and D. Vicari, Sapienza University of Rome, Italy (Last update: Apr, 2019)
function [P,Mh]=PMh_ROOTCLUS(SS,Res_S,P,Mh,W,Vh,offdiag,lambdal)
[N,J,H]=size(Mh);
ID=eye(J+1,J);
   for i=1:N
       t=zeros(J+1,1);
       for v=1:J+1
           P(i,:)=ID(v,:); 
           NP=sum(sum(P));     
           M=Mh;
               if v<J+1
                  M(i,:,:)=0;
               elseif v==J+1
                  [M(i,:,:)]=Mh_ROOTCLUS(P,Mh,i,Res_S,W,Vh,offdiag); 
               end
           for h=1:H
               if offdiag==1
                   t(v)=t(v)+ssq((Res_S(:,:,h)-P*W*P'-(P+M(:,:,h))*diag(Vh(h,:))*(P+M(:,:,h))').*(1-eye(N)));
               else
                   t(v)=t(v)+ssq(Res_S(:,:,h)-P*W*P'-(P+M(:,:,h))*diag(Vh(h,:))*(P+M(:,:,h))'); 
               end
           end 
           t(v)=t(v)/SS + lambdal*(N-NP)/(NP^2);
       end
       [~,pos]=min(t); 
       P(i,:)=ID(pos,:);
       if pos<J+1 
           Mh(i,:,:)=0; 
       elseif pos==J+1
           Mh(i,:,:)=M(i,:,:);
       end
   end 
end
%%%%%%

%%%%%%% Updating membership matrices Mh in PMh_ROOTCLUS  
% Written by L. Bocci and D. Vicari, Sapienza University of Rome, Italy (Last update: Apr, 2019)
function  [m]=Mh_ROOTCLUS(P,Mh,i,Res_S,W,Vh,offdiag)
 [N,J,H]=size(Mh); IDE=eye(J);
  Res_S=Res_S-repmat(P*W*P',[1 1 H]); 
  sP=sum(P);  sPfull=find(sP>0);
  
  if  sum(sP==0)==(J-1)
      Mh(i,:,:)=zeros(J,H);    
      Mh(i,sPfull,:)=ones(1,H); 
      m=Mh(i,:,:);
  else
           for h=1:H
               t=zeros(J,1); 
               for v=sPfull    
                   Mh(i,:,h)=IDE(v,:);
                   if offdiag==1
                       for hh=1:H
                         t(v)=t(v)+ssq(Res_S(:,:,hh)-(P*W*P'+(P+Mh(:,:,hh))*diag(Vh(hh,:))*(P+Mh(:,:,hh))').*(1-eye(N)));
                       end
                   elseif offdiag==0
                       for hh=1:H
                         t(v)=t(v)+ssq(Res_S(:,:,hh)-P*W*P'-(P+Mh(:,:,hh))*diag(Vh(hh,:))*(P+Mh(:,:,hh))');
                       end
                   end
               end
               t(sP==0)=inf;
               [~,pos]=min(t); 
               Mh(i,:,h)=IDE(pos,:);
           end
           m=Mh(i,:,:);
  end
end